==> libgii-0.1.tar.gz: libgii-0.1/gg/parse.c <==
/* $Id: parse.c,v 1.7 1998/12/11 20:44:35 marcus Exp $
******************************************************************************

   LibGG - Parsing code

   Copyright (C) 1998  Andrew Apted     [andrew.apted@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <ggi/gg.h>
#include "plat.h"

#define MAX_LINELEN	2048
#define COMMENT_CHAR	'#'

#define SKIPWHITE(str)	{while (isspace((int)*(str)) && *(str) != '\0' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}
#define SKIPNONWHITE(str) {while (!isspace((int)*(str)) && *(str) != '\0' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}
#define SKIPTOEND(str)	{while (*(str) != '\0' && *(str) != '\n' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}

int ggGetFileOpt(FILE *fp, const char **optnames, char **results, int ressize)
{
	char line[MAX_LINELEN];

	while (fgets(line, MAX_LINELEN, fp) != NULL) {
		int i;
		for (i = 0; optnames[i] != NULL; i++) {
			char *tmpptr, *endptr, *lineptr = line;
			int len = strlen(optnames[i]);
			
			SKIPWHITE(lineptr);
			if (strncasecmp(lineptr, optnames[i], len) != 0) {
				continue;
			}
			lineptr += len;
			if (! isspace((int)*lineptr)) continue;
			SKIPWHITE(lineptr);
			if (*lineptr == '\0' || *lineptr == COMMENT_CHAR) {
				continue;
			}

			endptr = tmpptr = lineptr;
			while (*tmpptr != '\0' && *tmpptr != COMMENT_CHAR) {
				SKIPNONWHITE(tmpptr);
				endptr = tmpptr;
				SKIPWHITE(tmpptr);
			}

			*endptr = '\0';
			if ((endptr - lineptr) + 1 > ressize) {
				return GGI_ENOMEM;
			}
			strcpy(results[i], lineptr);
			
			return i;
		}
	}

	return GGI_ENOTFOUND;
}


/*
******************************************************************************

   char *
   ggParseTarget(char *str, char *target, int max);
 
 	Parses a target descriptor out of a string, and returns a
 	pointer to after the last character of the target.  Handles
 	bracketized forms appropriately.  Returns NULL if an error
 	occurs.
 
 	For example:
 
               "display-multi:(display-kgi:/dev/head1):display-x\0"
                              ^                       ^
                              |__str                  |__result
 
               target: "display-kgi:/dev/head1"
	       
******************************************************************************
*/

char *ggParseTarget(char *str, char *target, int max)
{
	int bracketized = 0;
	int bracket_count = 0;

	while (*str != '\0' && isspace((int)*str)) {
		str++;
	}

	if (*str == '\0') {
		*target = '\0';
		return str;
	}

	if (*str == '(') {
		bracketized=1;
		bracket_count++;
		str++;
	}

	while (*str != '\0') {
		if (*str == '(') {
			bracket_count++;
		} else if (*str == ')') {
		
			if (bracket_count == 0) {
				fprintf(stderr, "libgg: Bad target "
					"descriptor : unexpected ')'\n");
				*target = 0;
				return NULL;
			}

			bracket_count--;

			if (bracketized && (bracket_count == 0)) {
				str++;
				break;
			}
		}

		if (max <= 2) {
			fprintf(stderr, "libgg: target descriptor "
				"too long\n");
			*target = '\0';
			return NULL;
		}
		
		*target++ = *str++; max--;
	}

	*target = '\0';

	if (bracket_count > 0) {
		fprintf(stderr, "libgg: Bad target descriptor : "
			"missing ')'\n");
		return NULL;
	}

	return str;
}

/*
******************************************************************************

   char *
   ggParseOptions(char *str, gg_option *optlist, int count);

 	Parses the target options out of the argument string.  Each
 	option begins with a '-' characters, and can be given a value
 	using the '=' characters.  Some examples "-nodb", "-parent=8".
 	Options are separated by ':' characters or whitespace.  
 
 	The recognizable options are stored in the array of gg_options:
 	`optlist'.  Matching is case sensitive, but options can be
 	abbreviated right down to a single letter.  Returns NULL if an
 	error occurred.
 	
	For example:
 
               "trueemu:-parent=8 -d=4:display-fbdev"
                        ^              ^
                        |__str         |__result
 	
 		optlist[0].name: "parent"     (in)
 		optlist11].name: "dither"     (in)
 
 		optlist[0].result: "8"        (out)
 		optlist11].result: "4"        (out)

******************************************************************************
*/

#define TERMINATOR(c)  (((c) == '\0') || isspace((int)c) || ((c) == ':'))

char *ggParseOptions(char *str, gg_option *optlist, int count)
{
	char name[MAX_OPTION_NAME]; int len;

	gg_option *cur; int i;


	for (;;) {

		while (*str != '\0' && isspace((int)*str)) 
			str++;

		if (*str != '-')
			break;

		/* parse option's name */

		len = 0;

		for (str++; ! TERMINATOR(*str) && (*str != '='); str++) {
			if (len < MAX_OPTION_NAME-1) {
				name[len++] = *str;
			}
		}

		if (len == 0) {
			fprintf(stderr, "libgg: Bad target options : "
				"missing option name\n");
			return NULL;
		}

		name[len] = '\0';

		/* EXPERIMENTAL FEATURE: shows config */

		if (strcmp(name, "showconfig") == 0) {
			fprintf(stderr, "libgg: CONFIG %d options.\n",
				count);
			for (i=0; i < count; i++) {
				fprintf(stderr, "libgg: CONFIG `%s' "
					"VALUE `%s'.\n", optlist[i].name, 
					optlist[i].result);
			}
			return NULL;
		}
		
		/* find option */

		cur = NULL;

		for (i=0; i < count; i++) {
			if (strncmp(optlist[i].name, name, len) == 0) {
				/* found it */
				cur = optlist + i;
			}
		}

		if (cur == NULL) {
			fprintf(stderr, "libgg: Unknown target option "
				"'%s'\n", name);
			return NULL;
		}

		/* store value */

		if (*str != '=') {
			strcpy(cur->result, "y");
			
		} else {
			len = 0;

			for (str++; ! TERMINATOR(*str); str++) {
				if (len < MAX_OPTION_RESULT-1) {
					cur->result[len++] = *str;
				}
			}
		}

		cur->result[len] = '\0';

		if (*str == ':')
			str++;
	}

	return str;
}

==> libgii_0.6.orig.tar.gz: libgii-0.6/gg/parse.c <==
/* $Id: parse.c,v 1.10 1999/02/25 15:00:51 marcus Exp $
******************************************************************************

   LibGG - Parsing code

   Copyright (C) 1998  Andrew Apted     [andrew.apted@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <ggi/gg.h>
#include "plat.h"

#define MAX_LINELEN	2048
#define COMMENT_CHAR	'#'
#define QUOTE_CHARS     "'\""

#define SKIPWHITE(str)	{while (isspace((int)*(str)) && *(str) != '\0' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}
#define SKIPNONWHITE(str) {while (!isspace((int)*(str)) && *(str) != '\0' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}
#define SKIPTOEND(str)	{while (*(str) != '\0' && *(str) != '\n' \
				&& *(str) != COMMENT_CHAR) { (str)++; }}

int ggGetFileOpt(FILE *fp, const char **optnames, char **results, int ressize)
{
	char line[MAX_LINELEN];

	while (fgets(line, MAX_LINELEN, fp) != NULL) {
		int i;
		for (i = 0; optnames[i] != NULL; i++) {
			char *tmpptr, *endptr, *lineptr = line;
			int len = strlen(optnames[i]);
			
			SKIPWHITE(lineptr);
			if (strncasecmp(lineptr, optnames[i], len) != 0) {
				continue;
			}
			lineptr += len;
			if (! isspace((int)*lineptr)) continue;
			SKIPWHITE(lineptr);
			if (*lineptr == '\0' || *lineptr == COMMENT_CHAR) {
				continue;
			}

			endptr = tmpptr = lineptr;
			while (*tmpptr != '\0' && *tmpptr != COMMENT_CHAR) {
				SKIPNONWHITE(tmpptr);
				endptr = tmpptr;
				SKIPWHITE(tmpptr);
			}

			*endptr = '\0';
			if ((endptr - lineptr) + 1 > ressize) {
				return GGI_ENOMEM;
			}
			strcpy(results[i], lineptr);
			
			return i;
		}
	}

	return GGI_ENOTFOUND;
}


/*
******************************************************************************

   char *
   ggParseTarget(const char *str, char *target, int max);
 
 	Parses a target descriptor out of a string, and returns a
 	pointer to after the last character of the target.  Handles
 	bracketized forms appropriately.  Returns NULL if an error
 	occurs.
 
 	For example:
 
               "display-multi:(display-kgi:/dev/head1):display-x\0"
                              ^                       ^
                              |__str                  |__result
 
               target: "display-kgi:/dev/head1"
	       
******************************************************************************
*/

char *ggParseTarget(const char *str, char *target, int max)
{
	int bracketized = 0;
	int bracket_count = 0;

	while (*str != '\0' && isspace((int)*str)) {
		str++;
	}

	if (*str == '\0') {
		*target = '\0';
		return (char *)str;
	}

	if (*str == '(') {
		bracketized=1;
		bracket_count++;
		str++;
	}

	while (*str != '\0') {
		if (*str == '(') {
			bracket_count++;
		} else if (*str == ')') {
		
			if (bracket_count == 0) {
				fprintf(stderr, "libgg: Bad target "
					"descriptor : unexpected ')'\n");
				*target = 0;
				return NULL;
			}

			bracket_count--;

			if (bracketized && (bracket_count == 0)) {
				str++;
				break;
			}
		}

		if (max <= 2) {
			fprintf(stderr, "libgg: target descriptor "
				"too long\n");
			*target = '\0';
			return NULL;
		}
		
		*target++ = *str++; max--;
	}

	*target = '\0';

	if (bracket_count > 0) {
		fprintf(stderr, "libgg: Bad target descriptor : "
			"missing ')'\n");
		return NULL;
	}

	return (char *)str;
}

/*
******************************************************************************

   char *
   ggParseOptions(const char *str, gg_option *optlist, int count);

 	Parses the target options out of the argument string.  Each
 	option begins with a '-' characters, and can be given a value
 	using the '=' characters.  Some examples "-nodb", "-parent=8".
 	Options are separated by ':' characters or whitespace.  
 
 	The recognizable options are stored in the array of gg_options:
 	`optlist'.  Matching is case sensitive, but options can be
 	abbreviated right down to a single letter.  Returns NULL if an
 	error occurred.
 	
	For example:
 
               "trueemu:-parent=8 -d=4:display-fbdev"
                        ^              ^
                        |__str         |__result
 	
 		optlist[0].name: "parent"     (in)
 		optlist11].name: "dither"     (in)
 
 		optlist[0].result: "8"        (out)
 		optlist11].result: "4"        (out)

******************************************************************************
*/

#define TERMINATOR(c)  (((c) == '\0') || isspace((int)c) || ((c) == ':'))

static inline const char *
ggParseOptionValue(const char *str, char *buf)
{
	int len = 0;
	int quote_c = 0;

	if ((*str != '\0') && (strchr(QUOTE_CHARS, *str) != NULL)) {
		/* get quote character */
		quote_c = *str++;
	}

	for (; *str != '\0'; str++) {

		if (quote_c) {
			if (*str == quote_c) {
				str++;
				break;
			}
		} else {
			if (TERMINATOR(*str)) {
				break;
			}
		}

		if ((*str == '\\') && (str[1] != '\0')) {
			/* handle escapes */
			str++;
		}

		if (len < MAX_OPTION_RESULT-1) {
			buf[len++] = *str;
		}
	}

	buf[len] = '\0';
	return str;
}

char *ggParseOptions(const char *str, gg_option *optlist, int count)
{
	char name[MAX_OPTION_NAME]; int len;

	gg_option *cur; int i;


	for (;;) {

		while (*str != '\0' && isspace((int)*str)) 
			str++;

		if (*str != '-')
			break;

		/* parse option's name */

		len = 0;

		for (str++; ! TERMINATOR(*str) && (*str != '='); str++) {
			if (len < MAX_OPTION_NAME-1) {
				name[len++] = *str;
			}
		}

		if (len == 0) {
			fprintf(stderr, "libgg: Bad target options : "
				"missing option name\n");
			return NULL;
		}

		name[len] = '\0';

		/* EXPERIMENTAL FEATURE: shows config */

		if (strcmp(name, "showconfig") == 0) {
			fprintf(stderr, "libgg: CONFIG has %d options%s\n",
				count, count ? ":" : ".");
			for (i=0; i < count; i++) {
				fprintf(stderr, "libgg: CONFIG option -%s "
					"= \"%s\".\n", optlist[i].name, 
					optlist[i].result);
			}
			return NULL;
		}
		
		/* find option */

		cur = NULL;

		for (i=0; i < count; i++) {
			if (strncmp(optlist[i].name, name, len) == 0) {
				/* found it */
				cur = optlist + i;
			}
		}

		if (cur == NULL) {
			fprintf(stderr, "libgg: Unknown target option "
				"'%s'\n", name);
			return NULL;
		}

		/* store value */

		if (*str != '=') {
			strcpy(cur->result, "y");
			
		} else {
			str = ggParseOptionValue(str+1, cur->result);
		}

		if (*str == ':')
			str++;
	}

	return (char *)str;
}
